use Test::More tests => 10;
use strict;

use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::Core::API::SystemCommand qw ( SystemCommand );
use DE_EPAGES::DataCache::API::Cache qw ( UpdateLocalConnection );
use DE_EPAGES::Database::API::Connection qw ( GetCurrentDBHandle GetTestSite RunOnStore );
use DE_EPAGES::Object::API::Factory qw (
    ExistsObjectByPath
    LoadObjectByPath
    LoadRootObject
);
use DE_EPAGES::Object::API::XML::ObjectDeleteHandler qw ( DeleteObjects );
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw ( ImportObjects );

use English;
use Readonly;
Readonly::Scalar my $DE_EPAGES => $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES';
Readonly::Scalar my $PERL      => $EXECUTABLE_NAME;
Readonly::Scalar my $ROOTPATH  => '/Test';

sub Main {
    ok( Init(), 'intialization done' );

    my $dbi = GetCurrentDBHandle();

    # import test shop including a price list 'RegularCustomer_EUR'
    ImportObjects('import.xml', $ROOTPATH);
    ok( 1, 'imported import.xml successfully' );

    # import a customer which is associated with 'RegularCustomer_EUR'
    SystemCommand([ $PERL, "$DE_EPAGES/Object/Scripts/import.pl",
        -storename => GetTestSite(),
        -path      => $ROOTPATH,
        'customers.xml'
    ]);  # ImportObjects('customers.xml', $ROOTPATH) does not produce the error
    ok( 1, 'imported customers.xml successfully' );

    # reset cache in order to get no outdated data
    UpdateLocalConnection($dbi->cacheKey);

    my $System = LoadRootObject();
    my $Shop = $System->folder('Test')->child('TestShop');
    my $Customers = $Shop->folder('Customers');
    ok( $Customers->existsChild('1001'), 'customer 1001 (mmustermann) imported' );

    my $Customer = $Customers->child('1001');
    my @PriceListAliases = map { $_->alias } @{ $Customer->get('PriceLists') };
    ok( /RegularCustomer_EUR/ ~~ @PriceListAliases, 'pricelist exists' );

    # delete a customer which is associated with 'RegularCustomer_EUR'
    DeleteObjects('customers.xml', $ROOTPATH);
    ok( 1, 'deleted customers.xml successfully' );

    # reset cache in order to get no outdated data
    UpdateLocalConnection($dbi->cacheKey);

    @PriceListAliases = map { $_->alias } @{ $Customer->get('PriceLists') };
    ok( !(/RegularCustomer_EUR/ ~~ @PriceListAliases), 'user removed from pricelist' );
    ok( $Customers->existsChild('1001'), 'customer 1001 (mmustermann) still exists' );
    ok( $Shop->folder('PriceLists')->existsChild('RegularCustomer_EUR'),
        'pricelist still exists on shop' );

    ok( Cleanup(), 'cleanup done' );

    return;
}

sub Init {
    Cleanup();

    my $System = LoadRootObject();
    $System->insertChild( 'Object', { Alias => 'Test' } );

    return 1;
}

sub Cleanup {
    LoadObjectByPath($ROOTPATH)->delete if ExistsObjectByPath($ROOTPATH);

    return 1;
}

RunScript(
    Debug   => 1,
    Message => 'long',
    Sub     => sub {
        RunOnStore(
            Store => GetTestSite(),
            Sub   => \&Main,
        );
    },
);

1;