use Test::More;
use strict;

use DE_EPAGES::Core::API::File qw ( DeleteFile );
use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::Customer::API::Shop qw ( GetNextCustomerNo GetCustomerFolder );
use DE_EPAGES::Database::API::Connection qw (
    GetCurrentDBHandle
    GetTestSite
    RunOnStore
    Transaction
);
use DE_EPAGES::Object::API::Factory qw ( LoadObjectByPath );
use DE_EPAGES::Order::API::Order qw ( CreateOrderWithContainer );
use DE_EPAGES::Shop::API::Shop qw ( GetNextNo );
use Storable qw ( lock_store lock_retrieve );

use Readonly;
Readonly::Array  my @CLASSNAMES => qw( Customer Invoice Product CustomerOrder );
Readonly::Scalar my $FILENAME   => 'GetNextNo.store';

local $ENV{'EPAGES_TESTSITE'} = 'Store';

sub TestGetNextNo {
    my ($ClassName, $FunctionRef) = @_;

    my @ForkedTasks;
    my $MaxNumbers = 10; ## no critic qw(ProhibitMagicNumbers)

    my $StorableObject = [];
    lock_store($StorableObject, $FILENAME);

    for (my $Count = 1; $Count <= 2; $Count++) {
        my $PID = fork();
        if ($PID) {  # parent
            push @ForkedTasks, $PID;
        } elsif ($PID == 0) {  # child
            my $NextNumbers = $FunctionRef->($Count, $MaxNumbers);

            ok( scalar @{ $NextNumbers } > 0,
                "new numbers were generated by child $Count ($ClassName)" );
            is( scalar @{ $NextNumbers }, $MaxNumbers,
                "$MaxNumbers new numbers were generated by child $Count ($ClassName)" );

            # use storable in order to keep these numbers outside this process
            $StorableObject = lock_retrieve($FILENAME);
            push @{ $StorableObject }, @{ $NextNumbers };
            lock_store($StorableObject, $FILENAME);
            exit 0;
        } else {
            die "Couldn't fork: $!\n";
        }
    }

    foreach my $Task ( @ForkedTasks ) {
        my $KID;
        do {
            $KID = waitpid($Task, 0);
        } while $KID > 0;
    }

    $StorableObject = lock_retrieve($FILENAME);
    my %Set = map { $_ => 1 } @{ $StorableObject };
    my $Result = (scalar @{ $StorableObject}) == (scalar keys %Set);
    ok( $Result, "bag of generated numbers from all processes is unique ($ClassName)");

    return $Result;
}

sub GetNextCustomerNoTask {
    my ($Count, $MaxNumbers) = @_;

    my @Numbers;
    RunOnStore(
        Store => GetTestSite(),
        Sub   => sub {
            my $Shop = LoadObjectByPath('/Shops/DemoShop');
            for (1..$MaxNumbers) {
                my $Number = GetNextNo($Shop, 'Customer');
                push @Numbers, $Number;
                sleep(rand(2));  # let the other child processes interfere
            }
        }
    );

    return \@Numbers;
}

sub GetNextInvoiceNoTask {
    my ($Count, $MaxNumbers) = @_;

    my @Numbers;
    RunOnStore(
        Store => GetTestSite(),
        Sub   => sub {
            my $Customer;
            my $Shop;

            Transaction(sub{
                $Shop = LoadObjectByPath( '/Shops/DemoShop' );
                my $CustomerNo = GetNextCustomerNo( $Shop ); # uses GetNextNo
                my $CustomerFolder = GetCustomerFolder( $Shop );
                $Customer = $Shop->get('CustomerClass')->insertObject({
                    Parent => $CustomerFolder,
                    Alias  => $CustomerNo,
                });
            });

            my $Order = CreateOrderWithContainer($Customer);
            for (1..$MaxNumbers) {
                my $Number = GetNextNo($Shop, 'Invoice', $Order);
                push @Numbers, $Number;
                sleep(rand(2));  # let the other child process interfere
            }
        }
    );

    return \@Numbers;
}

sub TestIncrementInteger {
    RunOnStore(
        Store => GetTestSite(),
        Sub   => sub {
            my $Shop = LoadObjectByPath( '/Shops/DemoShop' );
            my $Parent; # only needed for formatNumber

            # actually, we have to maintain the original state of each attribute
            # at test initialization. however, we need an extra database connection,
            # which again results in an issue where Mysql thinks AutoCommit isn't set
            my $dbh = GetCurrentDBHandle();
            my $sql = <<END_SQL;
UPDATE integerattribute
SET value = ?
WHERE objectid = ? AND attributeid = ?
END_SQL

            foreach my $ClassName ( @CLASSNAMES ) {
                # next number equals the last one
                my $IncNo = 0;
                my $Number1 = GetNextNo($Shop, $ClassName, $Parent, $IncNo);
                $Number1 =~ s/^0+//;
                my $Number2 = GetNextNo($Shop, $ClassName, $Parent, $IncNo);
                $Number2 =~ s/^0+//;
                is( $Number1, $Number2,
                    'next number equals the last one (' . $ClassName . ', IncNo=0)' );

                # next number > old one (if we assume no race conditions)
                $IncNo = 1;
                $Number2 = GetNextNo($Shop, $ClassName, $Parent, $IncNo);
                $Number2 =~ s/^0+//;
                ok( $Number2 > $Number1,
                    'next number is greater than before (' . $ClassName . ', IncNo=1)' );

                # reset to previous state (equals $Number1)
                my $AttributeName = 'Next' . $ClassName. 'No';
                $AttributeName = 'NextOrderNo' if $AttributeName eq 'NextCustomerOrderNo';
                my $NextNoAttribute = $Shop->class->attribute($AttributeName);
                $Number1 = 0 if $Number1 eq ''; # exception for Product
                $dbh->do($sql, [ $Number1, $Shop->id, $NextNoAttribute->id ]);
            }
        },
    );

    return;
}

sub Main {
    ok( Init(), 'intialization done' );

    SKIP: {
        # There seems to be an issue while using 'fork' in connection with Windows
        skip( 'Tests produce errors in Windows.', 4 ) if $^O eq 'MSWin32';

        my $Result = TestGetNextNo('Customer', \&GetNextCustomerNoTask);

        SKIP: {
            skip( 'GetNextNo is already broken', 2 ) unless $Result;

            TestGetNextNo('Invoice',  \&GetNextInvoiceNoTask);
        }
    }

    # actually, this test should come first, but if this subroutine is
    # called prior to TestGetNextNo, then an issue with the database connection
    # arises (Mysql complains that AutoCommit is not set, although it is set)
    TestIncrementInteger();

    ok( Cleanup(), 'cleanup done' );
    done_testing();

    return;
}

sub Init {
    Cleanup();

    return 1;
}

sub Cleanup {
    DeleteFile($FILENAME);

    return 1;
}

RunScript(
    Debug   => 1,
    Message => 'long',
    Sub     => \&Main,
);

1;